from dataclasses import dataclass

import outcome
import trio


@dataclass
class Future:
    """Container class for the results of an async function.
    """

    # TODO we can type this!
    result_chan: trio.abc.ReceiveChannel
    nursery: trio.Nursery  # TODO not sure if we need this but feels like good practice

    async def get(self):
        """Await the result of this :class:`Future` and unwrap the value. This will return a plain object
        if successful. If the bound function raised an exception, this will re-raise it.

        :return: The result of the bound unfunction, unwrapped
        :rtype: The return type of the bound function.
        """
        future_outcome = await self.outcome()
        return future_outcome.unwrap()

    async def outcome(self) -> outcome.Outcome:
        """Await the result of this :class:`Future` and return it as an instance of ``outcome.Outcome``,
        which can be one of ``Value`` or ``Error``. Can only be called once for this object instance.

        :raises RuntimeError: If this method is called more than once on an object instance.
        :return: Outcome of the function captured by this :class:`Future`.
        :rtype: outcome.Outcome
        """
        try:
            async with self.result_chan:
                return await self.result_chan.receive()
        except trio.ClosedResourceError:
            raise RuntimeError(
                "Trio resource closed (did you try to call outcome twice on this future?"
            )



from typing import List, Callable, Awaitable, Any

import outcome
import trio

from .future import Future


def run(nursery: trio.Nursery, async_fn: Callable[..., Awaitable[Any]], *args) -> Future:
    """Run an async function and capture its result in a Future. This is the main entrypoint for trio-future.

    Note that this is a synchronous function; it will immediately a :class:`Future` object. This object
    can be used to access the return value of ``async_fn``. However, the function will *not* have begun execution.
    Under the hood, we will pass the function to ``nursery.start_soon``; its execution will begin when we
    next defer to the scheduler.

    :param nursery: Nursery in which to run the function
    :type nursery: trio.Nursery
    :param async_fn: A trio-flavored async function to run. Positional arguments may be passed as trailing args, keyword arguments must use ``functools.partial`` or similar.
    :type async_fn: Callable[..., Awaitable[Any]]
    :return: A Future object allowing access to the return value of ``async_fn``.
    :rtype: Future
    """
    # Set buffer size to 1 so that producer can send a single result
    # witout blocking on a receiver.
    send_chan, recv_chan = trio.open_memory_channel(1)

    async def producer(*args):
        return_val = await outcome.acapture(async_fn, *args)
        # Shield sending the result from parent cancellation. This allows the Future to store
        # the outcome of the operation, namely that it was cancelled.
        # Note that the channel is buffered and only sent from here, so it should not block.
        with trio.CancelScope(shield=True):
            async with send_chan:
                await send_chan.send(return_val)

    nursery.start_soon(producer, *args)

    return Future(recv_chan, nursery)


def gather(nursery: trio.Nursery, futures: List[Future]) -> Future:
    """Concurrently run multiple Futures.

    This function will allow the provided Futures to run concurrently, and gather their results
    into a single :class:`Future` containing the results of all the inputs.
    That is, if each input Future contained an ``int``, then the returned Future will contain a ``List[int]``. In practice,
    the types need not be homogenous. The list enclosed in the returned Future will have the same ordering as the input list.

    If any Futures throw an exception, then the output Future will contain those exceptions, wrapped in a ``trio.MultiError``.

    :param nursery: Nursery that manages the concurrent execution of the provided futures
    :type nursery: trio.Nursery
    :param futures: Futures to run
    :type futures: List[Future]
    :return: A Future containing the results of all the provided futures.
    :rtype: Future
    """
    result_list = [None] * len(futures)
    parent_send_chan, parent_recv_chan = trio.open_memory_channel(0)
    child_send_chan, child_recv_chan = trio.open_memory_channel(0)

    async def producer():
        async with child_send_chan:
            for i in range(len(futures)):
                nursery.start_soon(child_producer, i, child_send_chan.clone())

    async def child_producer(i: int, out_chan):
        async with futures[i].result_chan:
            return_val = await futures[i].result_chan.receive()
            result_list[i] = return_val
            async with out_chan:
                await out_chan.send(i)

    async def receiver():
        async with child_recv_chan:
            async for i in child_recv_chan:
                # Just consume all results from the channel until exhausted
                pass
        # And then wrap up the result and push it to the parent channel
        errors = [e.error for e in result_list if isinstance(e, outcome.Error)]
        if len(errors) > 0:
            result = outcome.Error(trio.MultiError(errors))
        else:
            result = outcome.Value([o.unwrap() for o in result_list])
        async with parent_send_chan:
            await parent_send_chan.send(result)

    # Start parent producer, which will in turn start all children
    # (doing this inside the nursery because it needs to act async)
    nursery.start_soon(producer)
    nursery.start_soon(receiver)
    return Future(parent_recv_chan, nursery)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['trio_future']

package_data = \
{'': ['*']}

install_requires = \
['trio>=0.17.0,<0.18.0']

setup_kwargs = {
    'name': 'trio-future',
    'version': '0.1.1',
    'description': 'Capture the return values of concurrently executed trio functions',
    'long_description': '# trio-future\n\n## Overview\n\n`trio-future` allows you to capture the return values of concurrently executed trio functions. It\'s an altnerative to using trio channels to communicate results between tasks that feels more like programming with normal functions.\n\nConsider an example with this simple echo function:\n```python\nasync def echo(a: str) -> str:\n    await trio.sleep(0.5)\n    return a\n```\nWe can call our function and get its result back later when we are ready:\n```python\nasync with trio.open_nursery() as nursery:\n    # Call trio_future.run to synchronously get back a Future\n    future = trio_future.run(nursery, echo, "hello")\n    # When we call `await` and yield to scheduler, our function begins executing\n    await trio.sleep(0.1)\n    # We can `await` the function when we are ready\n    hello = await future.get() \n    # hello == "hello"\n```\nA common use-case is to run several tasks concurrently and wait for them all to complete. `trio-future` has a `gather` function like `asyncio.gather` to do this:\n```python\nasync with trio.open_nursery() as nursery:\n    fut_1 = run(nursery, echo, "hello")\n    fut_2 = run(nursery, echo, "world")\n    # Call `gather` to package the two Futures into a single Future object.\n    # Note that this is again a synchronous function.\n    joined_future = gather(nursery, [fut_1, fut_2])\n    # Again, when we `await` the result, we yield to the scheduler. This time, both\n    # of our futures will execute concurrently.\n    hello_world = await join_future.get()\n    # hello_world = ["hello", "world"]\n```\n',
    'author': 'Dan Frank',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/danielhfrank/trio-future',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
